# Copyright 2021 John Reese
# Licensed under the MIT License

from dataclasses import dataclass, field
from pathlib import Path
from shlex import quote
from typing import Any, Generator, List, Mapping, Optional, Sequence, Union

from packaging.version import Version


__all__ = [
    "Config",
    "ConfigError",
    "Context",
    "Job",
    "Options",
    "Result",
    "Version",
]

StrPath = Union[str, Path]


class ConfigError(ValueError):
    """Invalid configuration value"""


@dataclass
class Job:
    name: str
    run: Sequence[str]
    requires: Sequence[str] = ()
    once: bool = False
    parallel: bool = False
    isolated: bool = False

    def __post_init__(self) -> None:
        self.name = self.name.casefold()
        self.requires = tuple(r.casefold() for r in self.requires)


@dataclass
class Config:
    root: Path = field(default_factory=Path.cwd)
    jobs: Mapping[str, Job] = field(default_factory=dict)
    default: Sequence[str] = field(default_factory=list)
    values: Mapping[str, str] = field(default_factory=dict)
    versions: Sequence[Version] = field(default_factory=list)
    requirements: Sequence[str] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.default = tuple(d.casefold() for d in self.default)


@dataclass
class Context:
    python_version: Version
    python_path: Path
    venv: Path
    live: bool = False


@dataclass
class Options:
    config: Config = field(default_factory=Config)
    benchmark: bool = False
    debug: bool = False
    jobs: List[str] = field(default_factory=list)
    python: Optional[Version] = None
    clean: bool = False
    exit: bool = False


@dataclass
class CommandResult:
    exit_code: int
    stdout: str
    stderr: str

    @property
    def success(self) -> bool:
        return self.exit_code == 0

    @property
    def error(self) -> bool:
        return self.exit_code != 0


@dataclass
class Step:
    cmd: Sequence[str]
    job: Job
    context: Context
    config: Config

    def __await__(self) -> Generator[Any, None, "Result"]:
        return self.run().__await__()

    async def run(self) -> "Result":
        raise NotImplementedError


@dataclass
class Event:
    context: Context

    def __str__(self) -> str:
        return f"{self.context.python_version}> {self.__class__.__name__}"


@dataclass
class VenvCreate(Event):
    pass


@dataclass
class VenvReady(Event):
    pass


@dataclass
class JobEvent(Event):
    step: Step

    def __str__(self) -> str:
        cmd = " ".join(quote(arg) for arg in self.step.cmd)
        return f"{self.context.python_version} {self.step.job.name}> {cmd}"


@dataclass
class Start(JobEvent):
    pass


@dataclass
class Result(JobEvent, CommandResult):
    def __str__(self) -> str:
        cmd = " ".join(quote(arg) for arg in self.step.cmd)
        status = "OK" if self.success else "FAIL"
        return f"{self.context.python_version} {self.step.job.name}> {cmd} {status}"
